"""
Landmarker
"""

__version__ = "0.3.2"

__all__ = [
    "data",
    "datasets",
    "heatmap",
    "losses",
    "metrics",
    "models",
    "schedulers",
    "train",
    "uncertainty",
    "utils",
    "visualize",
]



"""Adaloss scheduler for sigma parameter of gaussian heatmaps"""

from collections import deque

import numpy as np


class AdalossScheduler:
    """
    Scheduler for sigma parameter of gaussian heatmaps. In essence the scheduler increases the
    problem difficulty by decreasing the sigma parameter. Implicitly by doing this, it creates a
    training curriculum.
        Proposed in "Adaloss: Adaptive Loss Function for Landmark Localization" - Teixeira et al.
            (2019)

    Args:
        nb_landmarks (int): number of landmarks
        rho (float, optional): [description]. Defaults to 0.9.
        window (int, optional): [description]. Defaults to 3.
        non_increasing (bool, optional): [description]. Defaults to False.
    """

    def __init__(
        self, nb_landmarks: int, rho: float = 0.9, window: int = 3, non_increasing: bool = False
    ):
        self.nb_landmarks = nb_landmarks
        self.rho = rho
        self.window = window
        self.non_increasing = non_increasing
        self.losses: list[deque[float]] = [deque([]) for _ in range(nb_landmarks)]
        self.prev_variances: list[float] = []
        self.counter = 0

    def __call__(self, losses, sigmas):
        self.counter += 1
        for i in range(self.nb_landmarks):
            self.losses[i].append(losses[i])
            if self.counter == self.window:
                self.prev_variances.append(
                    (1 / self.window) * (losses[i] - np.mean(self.losses[i])) ** 2
                )
            if self.counter > self.window:
                self.losses[i].popleft()
                variance = (1 / self.window) * (losses[i] - np.mean(self.losses[i])) ** 2
                adjustment = self.rho * (1 - self.prev_variances[i] / variance)
                if self.non_increasing:
                    adjustment = min(adjustment, 0.0)
                sigmas[i, 0] += adjustment
                sigmas[i, 1] += adjustment
                self.prev_variances[i] = variance



import numpy as np
import pytest
import torch

from src.landmarker.heatmap.decoder import coord_argmax
from src.landmarker.heatmap.generator import (
    GaussianHeatmapGenerator,
    LaplacianHeatmapGenerator,
    from_2by2_to_4by4,
    from_3by3_to_4by4,
)

landmarks = torch.tensor([[[10, 20], [30, 40], [50, 60]]], dtype=torch.int)
landmarks_batch = torch.tensor(
    [
        [[10, 20], [30, 40], [50, 60]],
        [[11, 22], [24, 35], [52, 58]],
        [[12, 24], [32, 42], [54, 60]],
        [[13, 26], [36, 46], [56, 61]],
    ],
    dtype=torch.int,
)


def test_from_2by2_to_4by4():
    affine_matrix = torch.tensor([[2.3, 0.0], [0.0, 0.5]])
    expected_result = torch.tensor(
        [[2.3, 0.0, 0.0, 0.0], [0.0, 0.5, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]
    )

    result = from_2by2_to_4by4(affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    affine_matrix = affine_matrix.unsqueeze(0)
    expected_result = expected_result.unsqueeze(0)

    result = from_2by2_to_4by4(affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    channel_affine_matrix = torch.tensor(
        [[[2.3, 0.0], [0.0, 0.5]], [[1, 0], [0, 1]], [[0, -1], [1, 0]]]
    )
    expected_result = torch.tensor(
        [
            [
                [2.3, 0.0, 0.0, 0.0],
                [0.0, 0.5, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            [
                [0.0, -1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
        ]
    )

    result = from_2by2_to_4by4(channel_affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    batch_channel_affine_matrix = channel_affine_matrix.unsqueeze(0).repeat(4, 1, 1, 1)
    expected_result = expected_result.unsqueeze(0).repeat(4, 1, 1, 1)

    result = from_2by2_to_4by4(batch_channel_affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)


def test_from_3by3_to_4by4():
    affine_matrix = torch.tensor([[2.3, 0.0, 0.0], [0.0, 0.5, 0.0], [0.0, 0.0, 1.0]])
    expected_result = torch.tensor(
        [[2.3, 0.0, 0.0, 0.0], [0.0, 0.5, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]
    )

    result = from_3by3_to_4by4(affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    affine_matrix = affine_matrix.unsqueeze(0)
    expected_result = expected_result.unsqueeze(0)

    result = from_3by3_to_4by4(affine_matrix)

    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    channel_affine_matrix = torch.tensor(
        [
            [[2.3, 0.0, 0.0], [0.0, 0.5, 0.0], [0.0, 0.0, 1.0]],
            [[1.0, 0.0, 0], [0, 1, 0], [0, 0, 1]],
            [[0, -1, 0], [1, 0, 0], [0, 0, 1]],
        ],
        dtype=float,
    )
    expected_result = torch.tensor(
        [
            [
                [2.3, 0.0, 0.0, 0.0],
                [0.0, 0.5, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            [
                [0.0, -1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
        ],
        dtype=float,
    )

    result = from_3by3_to_4by4(channel_affine_matrix)
    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)

    batch_channel_affine_matrix = channel_affine_matrix.unsqueeze(0).repeat(4, 1, 1, 1)
    expected_result = expected_result.unsqueeze(0).repeat(4, 1, 1, 1)
    result = from_3by3_to_4by4(batch_channel_affine_matrix)
    assert expected_result.shape == result.shape
    assert torch.allclose(result, expected_result)


def test_gaussian_heatmap_generator():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_gaussian_heatmap_generator_multiple_instances():
    landmarks_multiple_instances = torch.tensor(
        [
            [[[10, 20], [30, 40], [50, 60]], [[11, 22], [24, 35], [52, 58]]],
            [[[12, 24], [32, 42], [54, 60]], [[13, 26], [36, 46], [56, 61]]],
        ],
        dtype=torch.int,
    )

    assert landmarks_multiple_instances.shape == (2, 2, 3, 2)
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_multiple_instances.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks_multiple_instances)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 2, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for b in range(2):
        for c in range(2):
            for i in range(3):
                assert (
                    heatmaps[
                        b,
                        c,
                        landmarks_multiple_instances[b, c, i, 0],
                        landmarks_multiple_instances[b, c, i, 1],
                    ]
                    == 1
                )


def test_gaussian_heatmap_generator_multiple_instances_with_nan():
    landmarks_multiple_instances = torch.tensor(
        [
            [
                [[10, 20], [30, 40], [torch.nan, torch.nan]],
                [[11, 22], [24, 35], [torch.nan, torch.nan]],
            ],
            [
                [[12, 24], [32, 42], [torch.nan, torch.nan]],
                [[13, 26], [36, 46], [torch.nan, torch.nan]],
            ],
        ],
        dtype=torch.float,
    )

    assert landmarks_multiple_instances.shape == (2, 2, 3, 2)
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_multiple_instances.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks_multiple_instances)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 2, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for b in range(2):
        for c in range(2):
            for i in range(2):
                assert (
                    heatmaps[
                        b,
                        c,
                        int(landmarks_multiple_instances[b, c, i, 0]),
                        int(landmarks_multiple_instances[b, c, i, 1]),
                    ]
                    == 1
                )


def test_gaussian_heatmap_generator_learnable():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_gaussian_heatmap_generator_background():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
        background=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, landmarks.shape[1] + 1, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # check if background is first channel
    assert torch.equal(heatmaps[:, 0], 1 - heatmaps[:, 1:].sum(dim=1))

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j + 1, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_gaussian_heatmap_generator_all_points():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
        background=False,
        all_points=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, landmarks.shape[1] + 1, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check if all_points is first channel
    assert torch.equal(heatmaps[:, 0], heatmaps[:, 1:].sum(dim=1))

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j + 1, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_gaussian_heatmap_generator_not_continuous():
    landmarks_batch_ = landmarks_batch + torch.rand(landmarks_batch.shape)

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_batch_.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
        continuous=False,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch_)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert (
                heatmaps[
                    i,
                    j,
                    landmarks_batch_[i, j, 0].round().int(),
                    landmarks_batch_[i, j, 1].round().int(),
                ]
                == 1
            )


def test_gaussian_heatmap_generator_assymetric_sigmas():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            < heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_gaussian_heatmap_generator_rotation():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[0.5, 1], [0.5, 1], [1.5, 2.1]]),
        rotation=np.array([30, 45, 90]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the rotateed assymetry
    for i in range(4):
        # point 1: rotatation of 30°, point 2: 45°
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        # point 3: rotatation of 90°
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_gaussian_heatmap_generator_adaptive():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Change sigmas and rotation
    generator.set_sigmas(np.array([[0.5, 1], [0.5, 1], [1.5, 2.1]]))
    generator.set_rotation(np.array([30, 45, 90]))

    # Call generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the rotateed assymetry
    for i in range(4):
        # point 1: rotatation of 30°, point 2: 45°
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        # point 3: rotatation of 90°
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_gaussian_heatmap_generator_affine():
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Set affine matrix
    affine_matrix = torch.tensor(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=torch.float
    )

    # Call generator method
    heatmaps = generator(landmarks, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            < heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )

    # Set affine matrix (rotation of 90°)
    affine_matrix = torch.tensor(
        [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]], dtype=torch.float
    )
    # Call generator method
    heatmaps = generator(landmarks_batch, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            > heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )

    # Set affine matrix (rotation of 90°)
    affine_matrix = torch.tensor(
        [
            [[np.cos(np.pi / 6), -np.sin(np.pi / 6)], [np.sin(np.pi / 6), np.cos(np.pi / 6)]],
            [[np.cos(np.pi / 4), -np.sin(np.pi / 4)], [np.sin(np.pi / 4), np.cos(np.pi / 4)]],
            [[1.0, 0.0], [0.0, 1.0]],
            [[0.0, -1.0], [1.0, 0.0]],
        ],
        dtype=torch.float,
    )

    # Call generator method
    heatmaps = generator(landmarks_batch, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_gaussian_heatmap_generator_3d():
    landmarks_3d = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
        ],
        dtype=torch.float,
    )

    assert landmarks_3d.shape == (2, 3, 3)
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=1.0,
    )

    # Call generator method
    heatmaps = generator(landmarks_3d)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 3, 64, 64, 64)
    assert torch.all(heatmaps >= 0.0)  # No negative values
    sum_heatmaps = heatmaps.sum(dim=(2, 3, 4))
    assert torch.allclose(sum_heatmaps, torch.ones_like(sum_heatmaps))  # Because gamma is 1

    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    heatmaps = generator(landmarks_3d)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 3, 64, 64, 64)
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    for i in range(2):
        for j in range(3):
            assert heatmaps[
                i,
                j,
                landmarks_3d[i, j, 0].int(),
                landmarks_3d[i, j, 1].int(),
                landmarks_3d[i, j, 2].int(),
            ]


def test_gaussian_heatmap_generator_multiple_instances_3d():
    landmarks_3d = torch.tensor(
        [
            [
                [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
                [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            ],
            [
                [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
                [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
            ],
        ],
        dtype=torch.float,
    )

    assert landmarks_3d.shape == (2, 2, 3, 3)
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks_3d)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 2, 64, 64, 64)

    assert torch.all(heatmaps >= 0.0)  # No negative values

    for b in range(2):
        for c in range(2):
            for i in range(3):
                assert (
                    heatmaps[
                        b,
                        c,
                        landmarks_3d[b, c, i, 0].int(),
                        landmarks_3d[b, c, i, 1].int(),
                        landmarks_3d[b, c, i, 2].int(),
                    ]
                    == 1
                )


def test_gaussian_heatmap_generator_multiple_instances_with_nan_3d():
    landmarks_3d = torch.tensor(
        [
            [
                [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
                [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            ],
            [
                [[12, 22, 32], [22, 32, 42], [torch.nan, torch.nan, torch.nan]],
                [[13, 23, 33], [25, 41, 16], [torch.nan, torch.nan, torch.nan]],
            ],
        ],
        dtype=torch.float,
    )
    assert landmarks_3d.shape == (2, 2, 3, 3)
    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks_3d)

    # Check that the output has the correct shape
    assert heatmaps.shape == (2, 2, 64, 64, 64)

    assert torch.all(heatmaps >= 0.0)  # No negative values

    for b in range(2):
        for c in range(2):
            for i in range(2):
                if torch.isnan(landmarks_3d[b, c, i, 0]):
                    continue
                assert (
                    heatmaps[
                        b,
                        c,
                        landmarks_3d[b, c, i, 0].int(),
                        landmarks_3d[b, c, i, 1].int(),
                        landmarks_3d[b, c, i, 2].int(),
                    ]
                    == 1
                )


def test_gaussian_heatmap_generator_learnable_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.float,
    )

    assert landmarks_3d_batch.shape == (4, 3, 3)

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=True,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_3d_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(3):
            assert (
                heatmaps[
                    i,
                    j,
                    landmarks_3d_batch[i, j, 0].int(),
                    landmarks_3d_batch[i, j, 1].int(),
                    landmarks_3d_batch[i, j, 2].int(),
                ]
                == 1
            )


def test_gaussian_heatmap_generator_background_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.float,
    )

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
        background=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_3d_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 4, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(3):
            assert (
                heatmaps[
                    i,
                    j + 1,
                    landmarks_3d_batch[i, j, 0].int(),
                    landmarks_3d_batch[i, j, 1].int(),
                    landmarks_3d_batch[i, j, 2].int(),
                ]
                == 1
            )
        assert torch.allclose(heatmaps[i, 0], 1 - heatmaps[i, 1:].sum(dim=0))


def test_gaussian_heatmap_generator_all_points_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.float,
    )

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
        all_points=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_3d_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, landmarks_3d_batch.shape[1] + 1, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check if all_points is first channel
    assert torch.equal(heatmaps[:, 0], heatmaps[:, 1:].sum(dim=1))

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(3):
            assert (
                heatmaps[
                    i,
                    j + 1,
                    landmarks_3d_batch[i, j, 0].int(),
                    landmarks_3d_batch[i, j, 1].int(),
                    landmarks_3d_batch[i, j, 2].int(),
                ]
                == 1
            )


def test_gaussian_heatmap_generator_not_continuous_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.float,
    )

    landmarks_3d_batch_ = landmarks_3d_batch + torch.rand(landmarks_3d_batch.shape)

    assert landmarks_3d_batch.shape == (4, 3, 3)

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_3d_batch_)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    pred_landmarks_discrete = coord_argmax(heatmaps, spatial_dims=3)
    assert torch.equal(landmarks_3d_batch_.round(), pred_landmarks_discrete)


def test_gaussian_heatmap_generator_asymmetric_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.int32,
    )

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=np.array([[1, 0.5, 1], [0.5, 1, 1], [1.5, 2.1, 1]]),
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_3d_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(3):
            assert (
                heatmaps[
                    i,
                    j,
                    landmarks_3d_batch[i, j, 0].int(),
                    landmarks_3d_batch[i, j, 1].int(),
                    landmarks_3d_batch[i, j, 2].int(),
                ]
                == 1
            )

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[
                i,
                0,
                landmarks_3d_batch[i, 0, 0] + 1,
                landmarks_3d_batch[i, 0, 1],
                landmarks_3d_batch[i, 0, 2],
            ]
            > heatmaps[
                i,
                0,
                landmarks_3d_batch[i, 0, 0],
                landmarks_3d_batch[i, 0, 1] + 1,
                landmarks_3d_batch[i, 0, 2],
            ]
        )
        assert (
            heatmaps[
                i,
                1,
                landmarks_3d_batch[i, 1, 0] + 1,
                landmarks_3d_batch[i, 1, 1],
                landmarks_3d_batch[i, 1, 2],
            ]
            < heatmaps[
                i,
                1,
                landmarks_3d_batch[i, 1, 0],
                landmarks_3d_batch[i, 1, 1] + 1,
                landmarks_3d_batch[i, 1, 2],
            ]
        )
        assert (
            heatmaps[
                i,
                2,
                landmarks_3d_batch[i, 2, 0] + 1,
                landmarks_3d_batch[i, 2, 1],
                landmarks_3d_batch[i, 2, 2],
            ]
            < heatmaps[
                i,
                2,
                landmarks_3d_batch[i, 2, 0],
                landmarks_3d_batch[i, 2, 1] + 1,
                landmarks_3d_batch[i, 2, 2],
            ]
        )


def test_gaussian_heatmap_generator_affine_3d():
    landmarks_3d_batch = torch.tensor(
        [
            [[10, 20, 30], [20, 30, 40], [30, 15, 20]],
            [[11, 21, 31], [23, 39, 14], [20, 30, 30]],
            [[12, 22, 32], [22, 32, 42], [32, 17, 22]],
            [[13, 23, 33], [25, 41, 16], [22, 32, 32]],
        ],
        dtype=torch.int32,
    )

    # Create a generator
    generator = GaussianHeatmapGenerator(
        nb_landmarks=landmarks_3d_batch.shape[1],
        sigmas=np.array([[1, 0.5, 1], [0.5, 1, 1], [1.5, 2.1, 1]]),
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=None,
    )

    # Set affine matrix
    affine_matrix = torch.tensor(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=torch.float
    )

    # Call generator method
    heatmaps = generator(landmarks_3d_batch, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(3):
            assert (
                heatmaps[
                    i,
                    j,
                    landmarks_3d_batch[i, j, 0].int(),
                    landmarks_3d_batch[i, j, 1].int(),
                    landmarks_3d_batch[i, j, 2].int(),
                ]
                == 1
            )


def test_laplacian_heatmap_generator():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_laplacian_heatmap_generator_with_gamma():
    for gamma in range(1, 4):
        generator = LaplacianHeatmapGenerator(
            nb_landmarks=landmarks.shape[1],
            sigmas=1.0,
            heatmap_size=(64, 64),
            learnable=False,
            gamma=gamma,
        )

        # Call the generator method
        heatmaps = generator(landmarks)

        # Check that the output has the correct shape
        assert heatmaps.shape == (1, 3, 64, 64)

        # Check that the output values are within the expected range
        assert torch.all(heatmaps >= 0.0)  # No negative values
        assert torch.all(heatmaps <= (gamma / ((2 / 3) * torch.pi)))

        # Check that the output values are non-zero where expected
        for i in range(3):
            assert pytest.approx(heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]], 0.01) == (
                gamma / ((2 / 3) * torch.pi)
            )

        # Repeat same tests for batch
        # Call the generator method
        heatmaps = generator(landmarks_batch)

        # Check that the output has the correct shape
        assert heatmaps.shape == (4, 3, 64, 64)

        # Check that the output values are within the expected range
        assert torch.all(heatmaps >= 0.0)
        assert torch.all(heatmaps <= (gamma / ((2 / 3) * torch.pi)))

        # Check that the output values are non-zero where expected
        for i in range(4):
            for j in range(landmarks_batch.shape[1]):
                assert pytest.approx(
                    heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]], 0.01
                ) == (gamma / ((2 / 3) * torch.pi))
                assert pytest.approx(
                    heatmaps[i, j, landmarks_batch[i, j, 0] + 1, landmarks_batch[i, j, 1]], 0.01
                ) == (gamma / ((2 / 3) * torch.pi) * np.exp(-np.sqrt(3)))


def test_laplacian_heatmap_generator_subpixel():
    landmarks_batch_ = landmarks_batch + torch.rand(landmarks_batch.shape)

    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks_batch_.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch_)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    pred_landmarks_discrete = coord_argmax(heatmaps)
    assert torch.equal(landmarks_batch_.round(), pred_landmarks_discrete)


def test_laplacian_heatmap_generator_learnable():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_laplacian_heatmap_generator_background():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
        background=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, landmarks.shape[1] + 1, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # check if background is first channel
    assert torch.equal(heatmaps[:, 0], 1 - heatmaps[:, 1:].sum(dim=1))

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j + 1, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_laplacian_heatmap_generator_all_points():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=True,
        gamma=None,
        background=False,
        all_points=True,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, landmarks.shape[1] + 1, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check if all_points is first channel
    assert torch.equal(heatmaps[:, 0], heatmaps[:, 1:].sum(dim=1))

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j + 1, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1


def test_laplacian_heatmap_generator_not_continuous():
    landmarks_batch_ = landmarks_batch + torch.rand(landmarks_batch.shape)

    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks_batch_.shape[1],
        sigmas=1.0,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
        continuous=False,
    )

    # Call the generator method
    heatmaps = generator(landmarks_batch_)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert (
                heatmaps[
                    i,
                    j,
                    landmarks_batch_[i, j, 0].round().int(),
                    landmarks_batch_[i, j, 1].round().int(),
                ]
                == 1
            )


def test_laplacian_heatmap_generator_assymetric_sigmas():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            < heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_laplacian_heatmap_generator_rotation():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[0.5, 1], [0.5, 1], [1.5, 2.1]]),
        rotation=np.array([30, 45, 90]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Call generator method
    heatmaps = generator(landmarks)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the rotateed assymetry
    for i in range(4):
        # point 1: rotatation of 30°, point 2: 45°
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        # point 3: rotatation of 90°
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_laplacian_heatmap_generator_adaptive():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Change sigmas and rotation
    generator.set_sigmas(np.array([[0.5, 1], [0.5, 1], [1.5, 2.1]]))
    generator.set_rotation(np.array([30, 45, 90]))

    # Call generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the rotateed assymetry
    for i in range(4):
        # point 1: rotatation of 30°, point 2: 45°
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        # point 3: rotatation of 90°
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )


def test_laplacian_heatmap_generator_affine():
    # Create a generator
    generator = LaplacianHeatmapGenerator(
        nb_landmarks=landmarks.shape[1],
        sigmas=np.array([[1, 0.5], [0.5, 1], [1.5, 2.1]]),
        heatmap_size=(64, 64),
        learnable=False,
        gamma=None,
    )

    # Set affine matrix
    affine_matrix = torch.tensor(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=torch.float
    )

    # Call generator method
    heatmaps = generator(landmarks, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (1, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    # Because gamma = None and we don't deal with subpixel accuracy, we expect the heatmap to be 1 at the center of the
    # landmark.
    for i in range(3):
        assert heatmaps[0, i, landmarks[0, i, 0], landmarks[0, i, 1]] == 1

    # Call the generator method
    heatmaps = generator(landmarks_batch)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            > heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            < heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            < heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )

    # Set affine matrix (rotation of 90°)
    affine_matrix = torch.tensor(
        [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]], dtype=torch.float
    )
    # Call generator method
    heatmaps = generator(landmarks_batch, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # Check the assymetry
    for i in range(4):
        assert (
            heatmaps[i, 0, landmarks_batch[i, 0, 0] + 1, landmarks_batch[i, 0, 1]]
            < heatmaps[i, 0, landmarks_batch[i, 0, 0], landmarks_batch[i, 0, 1] + 1]
        )
        assert (
            heatmaps[i, 1, landmarks_batch[i, 1, 0] + 1, landmarks_batch[i, 1, 1]]
            > heatmaps[i, 1, landmarks_batch[i, 1, 0], landmarks_batch[i, 1, 1] + 1]
        )
        assert (
            heatmaps[i, 2, landmarks_batch[i, 2, 0] + 1, landmarks_batch[i, 2, 1]]
            > heatmaps[i, 2, landmarks_batch[i, 2, 0], landmarks_batch[i, 2, 1] + 1]
        )

    # Set affine matrix (rotation of 90°)
    affine_matrix = torch.tensor(
        [
            [[np.cos(np.pi / 6), -np.sin(np.pi / 6)], [np.sin(np.pi / 6), np.cos(np.pi / 6)]],
            [[np.cos(np.pi / 4), -np.sin(np.pi / 4)], [np.sin(np.pi / 4), np.cos(np.pi / 4)]],
            [[1.0, 0.0], [0.0, 1.0]],
            [[0.0, -1.0], [1.0, 0.0]],
        ],
        dtype=torch.float,
    )

    # Call generator method
    heatmaps = generator(landmarks_batch, affine_matrix=affine_matrix)

    # Check that the output has the correct shape
    assert heatmaps.shape == (4, 3, 64, 64)

    # Check that the output values are within the expected range
    assert torch.all(heatmaps >= 0.0)  # No negative values
    assert torch.all(heatmaps <= 1.0)  # Because gamma is None

    # Check that the output values are non-zero where expected
    for i in range(4):
        for j in range(landmarks_batch.shape[1]):
            assert heatmaps[i, j, landmarks_batch[i, j, 0], landmarks_batch[i, j, 1]] == 1

    # TODO add more elaborate tests for affine matrix



"""Stacked Cholesky Hourglass Network as proposed in "UGLLI Face Alignment: Estimating Uncertainty
with with Gaussian Log-Likelihood Loss" - Kumar et al. (2019)

#TODO: Check if the implementation is correct. The LUVLi paper seems to provide some code.
# see: https://github.com/abhi1kumar/LUVLi
"""

from typing import Sequence

import torch
from monai.networks.blocks import Convolution, ResidualUnit
from torch import nn

from landmarker.models.hourglass import Hourglass


class StackedCholeskyHourglass(nn.Module):
    """
    Stacked Cholesky Hourglass Network as proposed in "UGLLI Face Alignment: Estimating Uncertainty
    with Gaussian Log-Likelihood Loss" - Kumar et al. (2019).  It is a stack of hourglass networks
    with a Cholesky Estimator Network at the bottleneck of each hourglass. The output of the
    Cholesky Estimator Network is a lower triangular matrix that is used to estimate the covariance
    matrix of the Gaussian distribution of the predicted heatmaps. The covariance matrix is then
    used to compute the Gaussian Log-Likelihood Loss.

    Args:
        nb_stacks (int): number of hourglass networks to stack.
        img_size (tuple[int, int]): size of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        channels (Sequence[int], optional): number of output channels for each convolutional layer.
        conv_block (nn.Module, optional): convolutional block to use. Defaults to ResidualUnit.
        up_sample_mode (str, optional): upsampling mode. Defaults to 'nearest'.
    """

    def __init__(
        self,
        nb_stacks: int,
        img_size: tuple[int, int],
        in_channels: int,
        out_channels: int,
        channels: Sequence[int] = [64, 128, 256, 512],
        conv_block: nn.Module = ResidualUnit,  # type: ignore[assignment]
        up_sample_mode: str = "nearest",
    ) -> None:
        super().__init__()
        self.nb_stacks = nb_stacks
        self.stacks = nn.ModuleList(
            [
                CholeskyHourglass(
                    img_size=img_size,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    channels=channels,
                    conv_block=conv_block,
                    up_sample_mode=up_sample_mode,
                )
                for _ in range(nb_stacks)
            ]
        )
        self.output_to_feature_block = nn.ModuleList(
            [
                Convolution(
                    spatial_dims=len(img_size),
                    in_channels=out_channels,
                    out_channels=in_channels,
                    kernel_size=1,
                    adn_ordering="N",
                    norm="batch",
                )
                for _ in range(nb_stacks - 1)
            ]
        )

    def forward(self, x: torch.Tensor) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)

        Returns:
            heatmaps: list of predicted heatmaps of shape (batch_size, out_channels, *img_dims)
            cens: list of covariance matrices of the predicted heatmaps of shape (batch_size,
                out_channels, 2, 2)
        """
        heatmaps, cens = [], []
        next_x = x
        for i, stack in enumerate(self.stacks):
            pred, output_cen = stack(next_x)
            cens.append(output_cen)
            heatmaps.append(pred)
            if i < self.nb_stacks - 1:
                next_x = x + self.output_to_feature_block[i](pred)
        return heatmaps, cens


class CholeskyHourglass(nn.Module):
    """
    Proposed in "UGLLI Face Alignment: Estimating Uncertainty with
    Gaussian Log-Likelihood Loss" - Kumar et al. (2019)
    # TODO: Note that the implementation of Kumar et al. use DU-Net as the backbone.
    # We use the residual hourglass.

    Args:
        img_size (tuple[int, int]): size of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        channels (Sequence[int], optional): number of output channels for each convolutional layer.
        subunits (int, optional): number of subunits in each convolutional layer.
        up_sample_mode (str, optional): upsampling mode. Defaults to 'nearest'.

    Returns:
        pred: predicted heatmaps of shape (batch_size, out_channels, *img_dims)
        cen: covariance matrices of the predicted heatmaps of shape (batch_size,
            out_channels, 2, 2)
    """

    def __init__(
        self,
        img_size: tuple[int, int],
        in_channels: int,
        out_channels: int,
        channels: Sequence[int] = [64, 128, 256, 512],
        conv_block: nn.Module = ResidualUnit,  # type: ignore[assignment]
        up_sample_mode: str = "nearest",
    ) -> None:
        super().__init__()
        self.hg = Hourglass(
            spatial_dims=len(img_size),
            in_channels=in_channels,
            out_channels=out_channels,
            channels=channels,
            conv_block=conv_block,
            up_sample_mode=up_sample_mode,
        )
        self.img_size = img_size
        self.cen = CholeskyBlock(
            nb_classes=out_channels,
            input_units=channels[-1] * (img_size[0] // (2 ** len(channels))) ** 2,
            hidden_units=128,
            class_output_units=3,
        )

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)

        Returns:
            pred: predicted heatmaps of shape (batch_size, out_channels, *img_dims)
            cen: covariance matrices of the predicted heatmaps of shape (batch_size,
                out_channels, 2, 2)
        """
        down_outputs = []
        for down_block, skip_block in zip(self.hg.down_blocks, self.hg.skip_blocks):
            x = down_block(x)
            down_outputs.append(skip_block(x))
        x = self.hg.neck_block(x)
        cen_output = self.cen(x.flatten(1))
        for up_block, skip_output in zip(self.hg.up_blocks[::-1], down_outputs[::-1]):
            x = up_block(x + skip_output)
        x = self.hg.last_layer(x)
        return x, cen_output


class CholeskyBlock(nn.Module):
    """
    Cholesky Estimator Network tries to estimate the Cholesky decomposition of the covariance matrix
    out of the bottleneck of each hourglass of the stack.

    Proposed in "UGLLI Face Alignment: Estimating Uncertainty with
    Gaussian Log-Likelihood Loss" - Kumar et al. (2019)

    Args:
        nb_classes (int): number of output classes.
        input_units (int): number of input units.
        hidden_units (int): number of hidden units.
        class_output_units (int): number of output units.
        elu_alpha (float, optional): alpha parameter of the ELU activation function.
            Defaults to 1.0.
    """

    def __init__(
        self,
        nb_classes: int,
        input_units: int,
        hidden_units: int,
        class_output_units: int,
        elu_alpha: float = 1.0,
    ):
        super().__init__()
        self.nb_classes = nb_classes
        self.input_units = input_units
        self.hidden_units = hidden_units
        self.class_output_units = class_output_units
        self.seq = nn.Sequential(
            nn.Linear(input_units, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, nb_classes * class_output_units),
        )
        self.elu_alpha = elu_alpha
        self.elu = nn.ELU(alpha=elu_alpha)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, input_units)

        Returns:
            cov: covariance matrices of the predicted heatmaps of shape (batch_size,
                out_channels, 2, 2)
        """
        cholesky_decom = self.seq(x).view(-1, self.nb_classes, self.class_output_units)
        # Transform the output of the network to a lower triangular matrix
        l_mat = torch.zeros(
            (cholesky_decom.shape[0], self.nb_classes, 2, 2), device=cholesky_decom.device
        )
        l_mat[:, :, 0, 0] = self.elu(cholesky_decom[:, :, 0]) + self.elu_alpha
        l_mat[:, :, 1, 0] = cholesky_decom[:, :, 1]
        l_mat[:, :, 1, 1] = self.elu(cholesky_decom[:, :, 2]) + self.elu_alpha
        cov = l_mat @ l_mat.transpose(-2, -1)
        return cov



"""
Metrics for evaluating the performance of landmark detection models.
"""

from collections.abc import Iterable
from typing import Callable, Optional

import torch

from landmarker.utils.utils import pixel_to_unit


def point_error(
    true_landmarks: torch.Tensor,
    pred_landmarks: torch.Tensor,
    dim: Optional[tuple[int, ...] | torch.Tensor] = None,
    dim_orig: Optional[torch.Tensor] = None,
    pixel_spacing: Optional[torch.Tensor] = None,
    padding: Optional[torch.Tensor] = None,
    reduction: str = "mean",
):
    """
    Calculates the point error between true and predicted landmarks. The point error is the mean
    Euclidean distance between the true and predicted landmarks. If the pixel spacing is given, the
    point error is calculated in mm. If the pixel spacing is not given, the point error is
    calculated in pixels. If the dim and dim_orig are given, the landmarks are rescaled to the
    original image size.

    Args:
        true_landmarks (torch.Tensor): true landmarks
        pred_landmarks (torch.Tensor): predicted landmarks
        dim (Optional[tuple[int, ...] | torch.Tensor], optional): image size. Defaults to None.
        dim_orig (Optional[torch.Tensor], optional): original image size. Defaults to None.
        pixel_spacing (Optional[torch.Tensor], optional): pixel spacing. Defaults to None.
        padding (Optional[torch.Tensor], optional): padding. Defaults to None.
        reduction (str, optional): reduction method. Defaults to "mean". Can be "mean", "nanmean",
            or "none".
    """
    if pixel_spacing is None:
        pixel_spacing = torch.ones((len(true_landmarks), true_landmarks.shape[-1]))
    true_landmarks = pixel_to_unit(
        true_landmarks, pixel_spacing=pixel_spacing, dim=dim, dim_orig=dim_orig, padding=padding
    )
    pred_landmarks = pixel_to_unit(
        pred_landmarks, pixel_spacing=pixel_spacing, dim=dim, dim_orig=dim_orig, padding=padding
    )

    # Calculate MRE
    if reduction == "mean":
        return torch.mean(torch.sqrt(torch.sum((true_landmarks - pred_landmarks) ** 2, -1)))
    if reduction == "none":
        return torch.sqrt(torch.sum((true_landmarks - pred_landmarks) ** 2, -1))
    if reduction == "nanmean":
        return torch.nanmean(torch.sqrt(torch.sum((true_landmarks - pred_landmarks) ** 2, -1)))
    raise ValueError(f"Reduction method {reduction} not supported.")


def multi_instance_point_error(
    true_landmarks: torch.Tensor,
    pred_landmarks: list | torch.Tensor,
    dim: Optional[tuple[int, ...] | torch.Tensor | list[tuple[int, ...]]] = None,
    dim_orig: Optional[torch.Tensor] = None,
    pixel_spacing: Optional[torch.Tensor] = None,
    padding: Optional[torch.Tensor] = None,
    reduction: str = "mean",
    threshold: Optional[float] = None,
) -> (
    tuple[float, float, float, float, torch.Tensor]
    | tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
):
    """
    Calculates the multi-instance point error between true and predicted landmarks. The multi-
    instance point error is the mean Euclidean distance between the true and predicted landmarks
    for each instance. If the pixel spacing is given, the point error is calculated in mm. If the
    pixel spacing is not given, the point error is calculated in pixels. If the dim and dim_orig
    are given, the landmarks are rescaled to the original image size. The multi-instance point
    error is the mean of the instance point errors. Next to the multi-instance point error, the
    true positive and false positive rates are returned.

    Args:
        true_landmarks (list | torch.Tensor): true landmarks
        pred_landmarks (list | torch.Tensor): predicted landmarks
        dim (Optional[tuple[int, ...] | torch.Tensor | list[tuple[int, ...]]], optional): image
            size. Defaults to None.
        dim_orig (Optional[torch.Tensor], optional): original image size. Defaults to None.
        pixel_spacing (Optional[torch.Tensor], optional): pixel spacing. Defaults to None.
        padding (Optional[torch.Tensor], optional): padding. Defaults to None.
        reduction (str, optional): reduction method. Defaults to "mean". Can be "mean" or "none".

    Returns:
        tuple[float, float, float, float, torch.Tensor] | tuple[torch.Tensor, torch.Tensor,
            torch.Tensor, torch.Tensor, torch.Tensor]: multi-instance point error, true positive,
            false positive, false negative, predicted landmarks
    """
    if isinstance(pred_landmarks, torch.Tensor):
        pred_landmarks = transform_multi_instance_tensor(pred_landmarks)
    if isinstance(pred_landmarks[0], torch.Tensor):
        pred_landmarks = [[pred_landmarks]]
    if isinstance(pred_landmarks[0][0], torch.Tensor):
        pred_landmarks = [pred_landmarks]  # type: ignore
    assert len(true_landmarks) == len(pred_landmarks)
    assert len(true_landmarks[0]) == len(pred_landmarks[0])

    if dim is None:
        dim = [None] * len(true_landmarks)  # type: ignore
    if dim_orig is None:
        dim_orig = [None] * len(true_landmarks)  # type: ignore
    if pixel_spacing is None:
        pixel_spacing = [None] * len(true_landmarks)  # type: ignore
    if padding is None:
        padding = [None] * len(true_landmarks)  # type: ignore

    if isinstance(dim, tuple):
        dim = [dim] * len(true_landmarks)
    if isinstance(dim_orig, torch.Tensor) and len(dim_orig.shape) == 1:
        dim_orig = dim_orig.unsqueeze(0).repeat(len(true_landmarks), 1)
    if isinstance(pixel_spacing, torch.Tensor) and len(pixel_spacing.shape) == 1:
        pixel_spacing = pixel_spacing.unsqueeze(0).repeat(len(true_landmarks), 1)
    if isinstance(padding, torch.Tensor) and len(padding.shape) == 1:
        padding = padding.unsqueeze(0).repeat(len(true_landmarks), 1)

    pe = torch.zeros((len(true_landmarks), len(true_landmarks[0])))
    pe[pe == 0] = float("nan")
    tp = torch.zeros((len(true_landmarks), len(true_landmarks[0])))
    fp = torch.zeros((len(true_landmarks), len(true_landmarks[0])))
    fn = torch.zeros((len(true_landmarks), len(true_landmarks[0])))
    pred_landmarks_torch = torch.zeros_like(true_landmarks)
    pred_landmarks_torch[pred_landmarks_torch == 0] = float("nan")
    for b, landmark_batch in enumerate(true_landmarks):
        for c, landmark_class in enumerate(landmark_batch):
            picked_true = torch.zeros(len(true_landmarks[b][c]))
            for i, pred_landmark_instance in enumerate(pred_landmarks[b][c]):
                best_pe = float("inf")
                best_j = -1
                for j, true_landmark_instance in enumerate(landmark_class):
                    if picked_true[j]:
                        continue
                    assert pred_landmark_instance.shape == true_landmark_instance.shape
                    pe_inst = point_error(
                        true_landmark_instance.unsqueeze(0),
                        pred_landmark_instance.unsqueeze(0),
                        dim=dim[b],
                        dim_orig=dim_orig[b],  # type: ignore
                        pixel_spacing=pixel_spacing[b],  # type: ignore
                        padding=padding[b],  # type: ignore
                    )  # type: ignore
                    if pe_inst < best_pe:
                        best_pe = pe_inst
                        best_j = j
                if best_j != -1:
                    picked_true[best_j] = 1
                    if pe[b, c].isnan():
                        pe[b, c] = 0
                    pe[b, c] += best_pe
                    pred_landmarks_torch[
                        b,
                        c,
                        best_j,
                    ] = pred_landmark_instance
            if len(pred_landmarks[b][c]) > 0:
                pe[b, c] /= min(len(pred_landmarks[b][c]), len(landmark_class))
            tp[b, c] = torch.nansum(picked_true).item()
            fn[b, c] = len(true_landmarks[b][c]) - tp[b, c]
            fp[b, c] = max(len(pred_landmarks[b][c]) - torch.nansum(picked_true).item(), 0)
    if reduction == "mean":
        return (
            torch.nanmean(pe).item(),
            torch.sum(tp).item(),
            torch.sum(fp).item(),
            torch.sum(fn).item(),
            pred_landmarks_torch,
        )
    return pe, tp, fp, fn, pred_landmarks_torch


def transform_multi_instance_tensor(
    landmarks: torch.Tensor,
) -> list[torch.Tensor] | list[list[torch.Tensor]] | list[list[list[torch.Tensor]]] | list:
    """
    Transforms a multi-instance landmarks into a lists of landmarks. Where nan values are not
        included, since they are not landmarks.

    Args:
        tensor (torch.Tensor): multi-instance tensor

    Returns:
        list[torch.Tensor] | list[list[torch.Tensor]] | list[list[list[torch.Tensor]]]: list of
            tensors
    """
    if len(landmarks.shape) > 4 or len(landmarks.shape) == 1:
        raise ValueError(
            "If true_landmarks is a torch.Tensor, it must have at most 4 "
            + "dimensions. and at least 2 dimensions."
        )
    landmarks_list: (
        list[torch.Tensor] | list[list[torch.Tensor]] | list[list[list[torch.Tensor]]] | list
    ) = []
    for i in range(landmarks.shape[0]):
        if len(landmarks.shape) == 2:
            landmarks_list.append(landmarks[i])  # type: ignore
            continue
        landmarks_list.append([])  # type: ignore
        for j in range(landmarks.shape[1]):
            if len(landmarks.shape) == 3:
                if not torch.any(landmarks[i, j].isnan()):
                    landmarks_list[i].append(landmarks[i, j])  # type: ignore
                continue
            landmarks_list[i].append([])  # type: ignore
            for k in range(landmarks.shape[2]):
                if not torch.any(landmarks[i, j, k].isnan()):
                    landmarks_list[i][j].append(landmarks[i, j, k])  # type: ignore
    return landmarks_list


def sdr(
    radius: Iterable | float,
    true_landmarks: torch.Tensor,
    pred_landmarks: torch.Tensor,
    dim: Optional[tuple[int, ...] | torch.Tensor] = None,
    dim_orig: Optional[torch.Tensor] = None,
    pixel_spacing: Optional[torch.Tensor] = None,
    padding: Optional[torch.Tensor] = None,
    nanmean: bool = False,
) -> float | dict[float, float]:
    """
    Calculates the success detection rate (SDR), which is the percentage of
    predicted landmarks with a point error smaller or equal than the radius.

    Args:
        radius (Iterable | float): radius or list of radii
        true_landmarks (torch.Tensor): true landmarks
        pred_landmarks (torch.Tensor): predicted landmarks
        dim (Optional[tuple[int, ...] | torch.Tensor], optional): image size. Defaults to None.
        dim_orig (Optional[torch.Tensor], optional): original image size. Defaults to None.
        pixel_spacing (Optional[torch.Tensor], optional): pixel spacing. Defaults to None.
        padding (Optional[torch.Tensor], optional): padding. Defaults to None.
        nanmean (bool, optional): use nanmean instead of mean. Defaults to False.

    Returns:
        float | dict[float, float]: SDR or dictionary of SDRs for each radius
    """
    point_error_ = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=pixel_spacing,
        padding=padding,
        reduction="none",
    )
    if nanmean:
        reduce_func: Callable = torch.nanmean
    else:
        reduce_func = torch.mean
    if isinstance(radius, Iterable):
        return {r: reduce_func((point_error_ <= r).float()).item() * 100.0 for r in radius}
    return (reduce_func((point_error_ <= radius).float()) * 100).item()



"""Test the visualize module."""

import os

import cv2
import numpy as np
import pytest
from matplotlib import pyplot as plt

from landmarker.heatmap.generator import GaussianHeatmapGenerator
from src.landmarker.data import LandmarkDataset
from src.landmarker.data.landmark_dataset import HeatmapDataset, MaskDataset
from src.landmarker.models import ProbSpatialConfigurationNet
from src.landmarker.visualize import inspection_plot, prediction_inspect_plot

NSIM = 10  # number of simulated images


@pytest.fixture(scope="session", autouse=True)
def setup_data():
    """Create data for tests."""
    # Check if data folder exists and create it if not
    if not os.path.exists("tests/data"):
        os.makedirs("tests/data")
    all_landmarks = []
    all_img_uint8 = []
    for i in range(NSIM):
        img_uint8 = (np.random.rand(64, 64) * 256).astype(np.uint8)
        img_uint16 = (np.random.random((64, 64)) * 65535).astype(np.uint16)
        img_color = (np.random.rand(64, 64, 3) * 256).astype(np.uint8)
        landmarks = np.random.rand(8, 2)
        landmarks = landmarks * 64
        cv2.imwrite(f"tests/data/visualize_{i}_img_uint8.png", img_uint8)
        cv2.imwrite(
            f"tests/data/visualize_{i}_img_uint16.png",
            img_uint16,
        )
        cv2.imwrite(f"tests/data/visualize_{i}_img_color.png", img_color)
        all_landmarks.append(landmarks)
        all_img_uint8.append(img_uint8.reshape((1, 64, 64)))

    all_landmarks = np.stack(all_landmarks)

    np.save("tests/data/visualize_landmarks.npy", all_landmarks)
    np.save("visualize_all_img_uint8.npy", all_img_uint8)

    # define global variables
    pytest.landmarks = all_landmarks
    pytest.all_img_uint8 = all_img_uint8

    def clear_data():
        """Remove data for tests."""
        for i in range(NSIM):
            os.remove(f"tests/data/visualize_{i}_img_uint8.png")
            os.remove(f"tests/data/visualize_{i}_img_uint16.png")
            os.remove(f"tests/data/visualize_{i}_img_color.png")
        os.remove("tests/data/visualize_landmarks.npy")
        os.remove("visualize_all_img_uint8.npy")

    yield
    clear_data()


def test_inspection_plot():
    ds = LandmarkDataset(
        imgs=[f"tests/data/visualize_{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        class_names=[f"class_{i}" for i in range(8)],
        dim_img=(32, 32),
    )
    ds_heatmap = HeatmapDataset(
        imgs=[f"tests/data/visualize_{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        class_names=[f"class_{i}" for i in range(8)],
        dim_img=(32, 32),
    )

    ds_mask = MaskDataset(
        imgs=[f"tests/data/visualize_{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        class_names=[f"class_{i}" for i in range(8)],
        dim_img=(32, 32),
    )

    generator = GaussianHeatmapGenerator(ds.nb_landmarks, heatmap_size=(32, 32))

    idx = [2, 3]
    plt.ion()
    inspection_plot(ds, idx=idx, heatmap_generator=None, save_path=None)
    inspection_plot(ds, idx=idx, heatmap_generator=generator, save_path=None)

    idxs = [2, 3, 4]
    inspection_plot(ds, idx=idxs, heatmap_generator=generator, save_path=None)

    inspection_plot(ds, idx=idx, heatmap_generator=generator, save_path="tests/inspection_plot.png")

    os.remove("tests/inspection_plot.png")

    inspection_plot(ds_heatmap, idx=idx, save_path=None)
    inspection_plot(ds_heatmap, idx=idxs, save_path=None)
    inspection_plot(ds_heatmap, idx=idx, save_path="tests/inspection_plot.png")
    os.remove("tests/inspection_plot.png")

    inspection_plot(ds_mask, idx=idx, save_path=None)
    inspection_plot(ds_mask, idx=idxs, save_path=None)
    inspection_plot(ds_mask, idx=idx, save_path="tests/inspection_plot.png")
    os.remove("tests/inspection_plot.png")
    plt.close("all")


def test_prediction_inspect_plot():
    ds = LandmarkDataset(
        imgs=[f"tests/data/visualize_{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        class_names=[f"class_{i}" for i in range(8)],
        dim_img=(32, 32),
    )

    model = ProbSpatialConfigurationNet(
        spatial_dims=2,
        in_channels=1,
        out_channels=8,
        sp_image_input=False,
        la_channels=(32, 32, 32),
        la_strides=(2, 2),
    )

    idx = [2, 3]
    plt.ion()
    prediction_inspect_plot(ds, idx=idx, model=model, save_path=None)
    prediction_inspect_plot(ds, idx=idx, model=model, save_path="tests/prediction_inspect_plot.png")
    os.remove("tests/prediction_inspect_plot.png")
    plt.close("all")



"""
From plants to landmarks dataset.

This dataset is a subset of the Plant Centroids dataset, which is a dataset of images of plants
with their corresponding masks indicating the plant stem. This is a multi-instance single-class
landmark dataset, where the landmarks are the centroids of the masks. The dataset is split into a
training and three testing sets (A_W5, B_W5, C_W8). The "images" are stacks of the RGB and NIR
images.

The dataset can be downloaded from http://plantcentroids.cs.uni-freiburg.de/
"""

import os
import pathlib
import zipfile

import cv2
import numpy as np
import opendatasets as od  # type: ignore
import torch
from tqdm import tqdm  # type: ignore

from landmarker.data import HeatmapDataset, LandmarkDataset, MaskDataset
from landmarker.utils import get_paths


def get_plant_centroids_dataset(path_dir: str):
    if not os.path.exists(path_dir + "/plantcentroids"):
        od.download("http://plantcentroids.cs.uni-freiburg.de/dataset/plantcentroids.zip", path_dir)
        with zipfile.ZipFile(path_dir + "/plantcentroids.zip", "r") as zip_ref:
            zip_ref.extractall(path_dir)
        os.rename(path_dir + "/plantcentroids_dataset", path_dir + "/plantcentroids")
        os.remove(path_dir + "/plantcentroids.zip")

    train_mask_paths = []
    for train_folder in ["G_W5", "H_W5", "I_W6", "J_W8", "K_W8"]:
        train_mask_paths += sorted(
            get_paths(path_dir + "/plantcentroids/" + train_folder + "/label_sparse", "png")
        )

    train_img_rgb_paths = []
    for train_folder in ["G_W5", "H_W5", "I_W6", "J_W8", "K_W8"]:
        train_img_rgb_paths += sorted(
            get_paths(path_dir + "/plantcentroids/" + train_folder + "/rgb_undistorted", "png")
        )

    train_img_nir_paths = []
    for train_folder in ["G_W5", "H_W5", "I_W6", "J_W8", "K_W8"]:
        train_img_nir_paths += sorted(
            get_paths(path_dir + "/plantcentroids/" + train_folder + "/nir", "png")
        )

    A_W4_mask_paths = sorted(get_paths(path_dir + "/plantcentroids/A_W4/label_sparse", "png"))
    A_W4_img_rgb_paths = sorted(get_paths(path_dir + "/plantcentroids/A_W4/rgb_undistorted", "png"))
    A_W4_img_nir_paths = sorted(get_paths(path_dir + "/plantcentroids/A_W4/nir", "png"))

    B_W5_mask_paths = sorted(get_paths(path_dir + "/plantcentroids/B_W5/label_sparse", "png"))
    B_W5_img_rgb_paths = sorted(get_paths(path_dir + "/plantcentroids/B_W5/rgb_undistorted", "png"))
    B_W5_img_nir_paths = sorted(get_paths(path_dir + "/plantcentroids/B_W5/nir", "png"))

    C_W8_mask_paths = sorted(get_paths(path_dir + "/plantcentroids/C_W8/label_sparse", "png"))
    C_W8_img_rgb_paths = sorted(get_paths(path_dir + "/plantcentroids/C_W8/rgb_undistorted", "png"))
    C_W8_img_nir_paths = sorted(get_paths(path_dir + "/plantcentroids/C_W8/nir", "png"))

    train_out_imgs = stack_norm(train_img_rgb_paths, train_img_nir_paths, store=True)
    A_out_imgs = stack_norm(A_W4_img_rgb_paths, A_W4_img_nir_paths, store=True)
    B_out_imgs = stack_norm(B_W5_img_rgb_paths, B_W5_img_nir_paths, store=True)
    C_out_imgs = stack_norm(C_W8_img_rgb_paths, C_W8_img_nir_paths, store=True)

    return (
        train_mask_paths,
        train_img_rgb_paths,
        train_out_imgs,
        A_W4_mask_paths,
        A_W4_img_rgb_paths,
        A_out_imgs,
        B_W5_mask_paths,
        B_W5_img_rgb_paths,
        B_out_imgs,
        C_W8_mask_paths,
        C_W8_img_rgb_paths,
        C_out_imgs,
    )


def get_plant_centroids_mask_datasets(
    path_dir, train_transform=None, inference_transform=None, **kwargs
):
    (
        train_mask_paths,
        train_img_rgb_paths,
        train_out_imgs,
        A_W4_mask_paths,
        A_W4_img_rgb_paths,
        A_out_imgs,
        B_W5_mask_paths,
        B_W5_img_rgb_paths,
        B_out_imgs,
        C_W8_mask_paths,
        C_W8_img_rgb_paths,
        C_out_imgs,
    ) = get_plant_centroids_dataset(path_dir)
    test_kwargs = kwargs.copy()
    return (
        MaskDataset(
            train_img_rgb_paths, mask_paths=train_mask_paths, transform=train_transform, **kwargs
        ),
        MaskDataset(
            A_W4_img_rgb_paths,
            mask_paths=A_W4_mask_paths,
            transform=inference_transform,
            **test_kwargs,
        ),
        MaskDataset(
            B_W5_img_rgb_paths,
            mask_paths=B_W5_mask_paths,
            transform=inference_transform,
            **test_kwargs,
        ),
        MaskDataset(
            C_W8_img_rgb_paths,
            mask_paths=C_W8_mask_paths,
            transform=inference_transform,
            **test_kwargs,
        ),
    )


def get_plant_centroids_landmark_datasets(
    path_dir, train_transform=None, inference_transform=None, **kwargs
):
    mask_kwargs = kwargs.copy()
    mask_kwargs["dim_img"] = None
    try:
        mask_kwargs.pop("store_imgs")
    except KeyError:
        pass
    test_kwargs = kwargs.copy()
    mask_ds_trian, mask_ds_A, mask_ds_B, mask_ds_C = get_plant_centroids_mask_datasets(
        path_dir, store_imgs=False, **mask_kwargs
    )
    return (
        LandmarkDataset(
            imgs=mask_ds_trian.img_paths,
            landmarks=mask_ds_trian.landmarks_original,
            transform=train_transform,
            **kwargs,
        ),
        LandmarkDataset(
            imgs=mask_ds_A.img_paths,
            landmarks=mask_ds_A.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
        LandmarkDataset(
            imgs=mask_ds_B.img_paths,
            landmarks=mask_ds_B.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
        LandmarkDataset(
            imgs=mask_ds_C.img_paths,
            landmarks=mask_ds_C.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
    )


def get_plant_centroids_heatmap_datasets(
    path_dir, train_transform=None, inference_transform=None, **kwargs
):
    mask_kwargs = kwargs.copy()
    mask_kwargs["dim_img"] = None
    mask_kwargs.pop("sigma")
    try:
        mask_kwargs.pop("store_imgs")
    except KeyError:
        pass
    test_kwargs = kwargs.copy()
    mask_ds_trian, mask_ds_A, mask_ds_B, mask_ds_C = get_plant_centroids_mask_datasets(
        path_dir, store_imgs=False, **mask_kwargs
    )
    return (
        HeatmapDataset(
            imgs=mask_ds_trian.img_paths,
            landmarks=mask_ds_trian.landmarks_original,
            transform=train_transform,
            **kwargs,
        ),
        HeatmapDataset(
            imgs=mask_ds_A.img_paths,
            landmarks=mask_ds_A.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
        HeatmapDataset(
            imgs=mask_ds_B.img_paths,
            landmarks=mask_ds_B.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
        HeatmapDataset(
            imgs=mask_ds_C.img_paths,
            landmarks=mask_ds_C.landmarks_original,
            transform=inference_transform,
            **test_kwargs,
        ),
    )


def stack_norm(img_paths, img_nir_paths, store=False):
    img_stack = []
    for i, img_path in enumerate(tqdm(img_paths)):
        if store:
            img_name = pathlib.Path(img_path).name.replace(".png", ".npy")
            img_stack_path = pathlib.Path(img_path).parents[1].joinpath(f"rgb_nir/{img_name}")
            img_stack.append(img_stack_path.__str__())
            if not os.path.exists(img_stack_path.parents[0]):
                os.makedirs(img_stack_path.parents[0])
            if os.path.exists(img_stack[i]):
                continue
        img = cv2.imread(img_path, cv2.IMREAD_UNCHANGED)
        img = img.transpose(2, 0, 1)
        if img.dtype == np.uint16:
            img = torch.tensor(img / 65535.0, dtype=float).view(-1, img.shape[-2], img.shape[-1])
        elif img.dtype == np.uint8:
            img = torch.tensor(img / 255.0, dtype=float).view(-1, img.shape[-2], img.shape[-1])
        else:
            raise ValueError(f"Image type {img.dtype} not supported")
        nir = cv2.imread(img_nir_paths[i], cv2.IMREAD_UNCHANGED)
        if len(nir.shape) == 3:
            nir = nir[:, :, 0]
        if nir.dtype == np.uint16:
            nir = torch.tensor(nir / 65535.0, dtype=float).view(-1, nir.shape[-2], nir.shape[-1])
        elif nir.dtype == np.uint8:
            nir = torch.tensor(nir / 255.0, dtype=float).view(-1, nir.shape[-2], nir.shape[-1])
        else:
            raise ValueError(f"Image type {nir.dtype} not supported")
        if store:
            np.save(
                img_stack[i],
                torch.cat((img, nir), dim=0).numpy(),
            )
        else:
            img_stack.append(torch.cat((img, nir), dim=0))
    return img_stack



"""Heatmap generator"""

from abc import abstractmethod
from typing import Optional

import numpy as np
import torch
from torch import nn


class HeatmapGenerator(nn.Module):
    """
    Heatmap generator abstract class for generating heatmaps from landmarks

    Args:
        nb_landmarks (int): number of landmarks
        sigmas (float or list[float] or torch.Tensor or np.ndarray): sigmas of the heatmap function
        gamma (float or None): scaling factor of the heatmap function
        rotation (float or list[float] or torch.Tensor or np.ndarray): rotation of the heatmap
            function
        heatmap_size (tuple[int, ...]): size of the heatmap
        learnable (bool): whether the sigmas and rotation are learnable
        background (bool): whether to add a background channel to the heatmap
        all_points (bool): whether to add a channel with the sum of all the landmarks
        continuous (bool): whether to use continuous or discrete landmarks
        na_zero (bool): whether to set the value of the landmarks to zero if they are not available
    """

    def __init__(
        self,
        nb_landmarks: int,
        sigmas: float | list[float] | torch.Tensor | np.ndarray = 1.0,
        gamma: Optional[float] = None,
        rotation: float | list[float] | torch.Tensor | np.ndarray = 0,
        heatmap_size: tuple[int, ...] = (512, 512),
        learnable: bool = False,
        background: bool = False,
        all_points: bool = False,
        continuous: bool = True,
        na_zero: bool = False,
    ) -> None:
        super(HeatmapGenerator, self).__init__()
        self.nb_landmarks = nb_landmarks
        self.learnable = learnable
        self.heatmap_size = heatmap_size
        self.spatial_dims = len(heatmap_size)
        self.set_sigmas(sigmas)
        self.set_rotation(rotation)
        self.gamma = gamma
        self.background = background
        self.all_points = all_points
        self.continuous = continuous
        self.na_zero = na_zero

    def set_sigmas(self, sigmas: float | list[float] | torch.Tensor | np.ndarray) -> None:
        """
        Set the sigmas of the heatmap function.

        Args:
            sigmas (float or list[float] or torch.Tensor or np.ndarray): sigmas of the heatmap
                function
        """
        if isinstance(sigmas, torch.Tensor):
            device = sigmas.device
        else:
            device = torch.device("cpu")
        sigmas = torch.tensor(
            sigmas, device=device, dtype=torch.float, requires_grad=self.learnable
        )
        sigmas = (
            torch.ones(
                (self.nb_landmarks, self.spatial_dims),
                requires_grad=self.learnable,
                device=device,
                dtype=torch.float,
            )
            * sigmas
        )
        if self.learnable:
            self.sigmas = torch.nn.Parameter(sigmas)
        else:
            self.register_buffer("sigmas", sigmas)

    def set_rotation(self, rotation: float | list[float] | torch.Tensor | np.ndarray) -> None:
        """
        Set the rotation of the heatmap function.

        Args:
            rotation (float or list[float] or torch.Tensor or np.ndarray): rotation of the heatmap
                function
        """
        if hasattr(self, "sigmas"):
            device = self.sigmas.device
        elif isinstance(rotation, torch.Tensor):
            device = rotation.device
        else:
            device = torch.device("cpu")
        rotation = torch.tensor(
            rotation, device=device, dtype=torch.float, requires_grad=self.learnable
        )
        if self.spatial_dims == 2:
            rotation = (
                torch.ones(
                    (self.nb_landmarks,),
                    requires_grad=self.learnable,
                    device=device,
                    dtype=torch.float,
                )
                * rotation
            )
        else:
            rotation = (
                torch.ones(
                    (self.nb_landmarks, self.spatial_dims),
                    requires_grad=self.learnable,
                    device=device,
                    dtype=torch.float,
                )
                * rotation
            )
        if self.learnable:
            self.rotation = torch.nn.Parameter(rotation)
        else:
            self.register_buffer("rotation", rotation)

    def __call__(
        self, landmarks: torch.Tensor, affine_matrix: torch.Tensor = torch.eye(4)
    ) -> torch.Tensor:
        affine_matrix = affine_matrix.to(landmarks.device)
        assert affine_matrix.shape[-1] == affine_matrix.shape[-2]
        if len(affine_matrix.shape) == 2:
            affine_matrix = affine_matrix.unsqueeze(0)
        if affine_matrix.shape[-1] == 2:
            # go from 2 by 2 affine matrix to 4 by 4
            affine_matrix = from_2by2_to_4by4(affine_matrix)
        elif affine_matrix.shape[-1] == 3:
            # go from 3 by 3 affine matrix to 4 by 4
            affine_matrix = from_3by3_to_4by4(affine_matrix)
        affine_matrix = affine_matrix.unsqueeze(1)
        heatmaps = torch.zeros(
            (landmarks.shape[0], landmarks.shape[1], *self.heatmap_size), device=self.sigmas.device
        )
        if len(landmarks.shape) == 2:
            heatmaps = self.create_heatmap(landmarks.unsqueeze(1), self.gamma, affine_matrix)
        elif len(landmarks.shape) == 3:
            heatmaps = self.create_heatmap(landmarks, self.gamma, affine_matrix)
        else:
            heatmaps = self.create_heatmap(landmarks, self.gamma, affine_matrix).nansum(dim=2)
        if self.all_points:
            heatmaps = torch.cat((heatmaps.sum(dim=1, keepdim=True), heatmaps), 1)
        elif self.background:
            heatmaps = torch.cat(
                (
                    torch.ones(
                        (heatmaps.shape[0], 1, *self.heatmap_size), device=self.sigmas.device
                    )
                    - heatmaps.sum(dim=1, keepdim=True),
                    heatmaps,
                ),
                1,
            )
        if self.na_zero:
            heatmaps[torch.isnan(heatmaps)] = 0
        return heatmaps

    @abstractmethod
    def heatmap_fun(
        self,
        landmark_t: torch.Tensor,
        coords: torch.Tensor,
        covariance: torch.Tensor,
        gamma: Optional[float],
    ):
        """Abstract heatmap function

        Args:
            landmark_t (torch.Tensor): coordinates of the landmark (y, x) or (y, x, z)
            coords (torch.Tensor): coordinates of the pixel (y, x) or (y, x, z)
            covariance (torch.Tensor): covariance matrix
            gamma (float or None): scaling factor of the heatmap function
        """

    def get_covariance_matrix(self, return4by4: bool = False) -> torch.Tensor:
        """
        Get the covariance matrix of the heatmap function.

        Args:
            return4by4 (bool): whether to return a 4 by 4 covariance matrix or a spatial_dims by
                spatial_dims covariance matrix.

        Returns:
            torch.Tensor: covariance matrix
        """
        if self.spatial_dims == 2:
            rotation = torch.stack(
                (
                    torch.stack((torch.cos(self.rotation), -torch.sin(self.rotation)), dim=-1),
                    torch.stack((torch.sin(self.rotation), torch.cos(self.rotation)), dim=-1),
                ),
                dim=-2,
            )

            diagonal = torch.diag_embed((self.sigmas**2))
            if return4by4:
                rotation = from_2by2_to_4by4(rotation)
                diagonal = from_2by2_to_4by4(diagonal)
        else:  # 3D case
            # see: https://msl.cs.uiuc.edu/planning/node102.html
            rotation_yaw = torch.stack(
                (
                    torch.stack(
                        (
                            torch.cos(self.rotation[..., 0]),
                            -torch.sin(self.rotation[..., 0]),
                            torch.zeros_like(self.rotation[..., 0]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            torch.sin(self.rotation[..., 0]),
                            torch.cos(self.rotation[..., 0]),
                            torch.zeros_like(self.rotation[..., 0]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            torch.zeros_like(self.rotation[..., 0]),
                            torch.zeros_like(self.rotation[..., 0]),
                            torch.ones_like(self.rotation[..., 0]),
                        ),
                        dim=-1,
                    ),
                ),
                dim=-2,
            )
            rotation_pitch = torch.stack(
                (
                    torch.stack(
                        (
                            torch.cos(self.rotation[..., 1]),
                            torch.zeros_like(self.rotation[..., 1]),
                            torch.sin(self.rotation[..., 1]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            torch.zeros_like(self.rotation[..., 1]),
                            torch.ones_like(self.rotation[..., 1]),
                            torch.zeros_like(self.rotation[..., 1]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            -torch.sin(self.rotation[..., 1]),
                            torch.zeros_like(self.rotation[..., 1]),
                            torch.cos(self.rotation[..., 1]),
                        ),
                        dim=-1,
                    ),
                ),
                dim=-2,
            )
            rotation_roll = torch.stack(
                (
                    torch.stack(
                        (
                            torch.ones_like(self.rotation[..., 2]),
                            torch.zeros_like(self.rotation[..., 2]),
                            torch.zeros_like(self.rotation[..., 2]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            torch.zeros_like(self.rotation[..., 2]),
                            torch.cos(self.rotation[..., 2]),
                            -torch.sin(self.rotation[..., 2]),
                        ),
                        dim=-1,
                    ),
                    torch.stack(
                        (
                            torch.zeros_like(self.rotation[..., 2]),
                            torch.sin(self.rotation[..., 2]),
                            torch.cos(self.rotation[..., 2]),
                        ),
                        dim=-1,
                    ),
                ),
                dim=-2,
            )
            rotation = rotation_yaw @ rotation_pitch @ rotation_roll

            diagonal = torch.diag_embed((self.sigmas**2))
            if return4by4:
                rotation = from_3by3_to_4by4(rotation)
                diagonal = from_3by3_to_4by4(diagonal)
        covariance = rotation @ diagonal @ rotation.transpose(-2, -1)
        return covariance

    def create_heatmap(
        self, landmarks: torch.Tensor, gamma: Optional[float], affine_matrix: torch.Tensor
    ) -> torch.Tensor:
        """Create a heatmap for a given landmark in an image returns a heatmap with the same size as
        the image. Works with batches and multiple landmarks.

        Args:
            landmarks (torch.Tensor): landmarks of shape (B, C, M, S) or (B, C, S)
            gamma (float or None): scaling factor of the heatmap function
            affine_matrix (torch.Tensor): affine matrix of shape (B, 1, 4, 4)

        Returns:
            torch.Tensor: heatmap of shape (B, C, M, H, W,) or (B, C, H, W) if landmarks are 2D
                else (B, C, M, D, H, W) or (B, C, D, H, W) if landmarks are 3D
        """
        covariance = self.get_covariance_matrix(return4by4=True)
        covariance = affine_matrix @ covariance @ affine_matrix.transpose(-2, -1)

        if self.spatial_dims == 2:
            x = landmarks[..., 1]
            y = landmarks[..., 0]
            x_round = torch.round(x).int()
            y_round = torch.round(y).int()
            xs = torch.arange(
                0, self.heatmap_size[1], 1, dtype=torch.float32, device=self.sigmas.device
            )
            ys = torch.arange(
                0, self.heatmap_size[0], 1, dtype=torch.float32, device=self.sigmas.device
            )
            xs, ys = torch.meshgrid(xs, ys, indexing="xy")
            pre_shape = tuple(1 for _ in range(len(landmarks.shape[:-1])))
            xs = xs.view(*pre_shape, *xs.shape).repeat(*landmarks.shape[:-1], 1, 1)
            ys = ys.view(*pre_shape, *ys.shape).repeat(*landmarks.shape[:-1], 1, 1)
            if self.continuous:
                x_t, y_t = x, y
            else:
                x_t, y_t = x_round, y_round
            x_t = x_t.view(*landmarks.shape[:-1], 1, 1)
            y_t = y_t.view(*landmarks.shape[:-1], 1, 1)
            heatmap = self.heatmap_fun(
                torch.stack((y_t, x_t), -1), torch.stack((ys, xs), -1), covariance, gamma
            )
        else:
            z = landmarks[..., 0]
            y = landmarks[..., 1]
            x = landmarks[..., 2]
            z_round = torch.round(z).int()
            y_round = torch.round(y).int()
            x_round = torch.round(x).int()
            zs = torch.arange(
                0, self.heatmap_size[0], 1, dtype=torch.float32, device=self.sigmas.device
            )
            ys = torch.arange(
                0, self.heatmap_size[1], 1, dtype=torch.float32, device=self.sigmas.device
            )
            xs = torch.arange(
                0, self.heatmap_size[2], 1, dtype=torch.float32, device=self.sigmas.device
            )
            zs, ys, xs = torch.meshgrid(zs, ys, xs, indexing="ij")
            pre_shape = tuple(1 for _ in range(len(landmarks.shape[:-1])))
            zs = zs.view(*pre_shape, *zs.shape).repeat(*landmarks.shape[:-1], 1, 1, 1)
            ys = ys.view(*pre_shape, *ys.shape).repeat(*landmarks.shape[:-1], 1, 1, 1)
            xs = xs.view(*pre_shape, *xs.shape).repeat(*landmarks.shape[:-1], 1, 1, 1)
            if self.continuous:
                z_t, y_t, x_t = z, y, x
            else:
                z_t, y_t, x_t = z_round, y_round, x_round
            z_t = z_t.view(*landmarks.shape[:-1], 1, 1, 1)
            y_t = y_t.view(*landmarks.shape[:-1], 1, 1, 1)
            x_t = x_t.view(*landmarks.shape[:-1], 1, 1, 1)
            heatmap = self.heatmap_fun(
                torch.stack((z_t, y_t, x_t), -1), torch.stack((zs, ys, xs), -1), covariance, gamma
            )
        return heatmap


class GaussianHeatmapGenerator(HeatmapGenerator):
    """
    Gaussian heatmap generator for generating heatmaps from landmarks.

    Args:
        nb_landmarks (int): number of landmarks
        sigmas (float or list[float] or torch.Tensor or np.ndarray): sigmas of the gaussian heatmap
            function
        gamma (float or None): scaling factor of the gaussian heatmap function
        rotation (float or list[float] or torch.Tensor or np.ndarray): rotation of the gaussian
            heatmap function
        heatmap_size (tuple[int, int]): size of the heatmap
        learnable (bool): whether the sigmas and rotation are learnable
        background (bool): whether to add a background channel to the heatmap
        all_points (bool): whether to add a channel with the sum of all the landmarks
        continuous (bool): whether to use continuous or discrete landmarks
        na_zero (bool): whether to set the value of the landmarks to zero if they are not available
    """

    def __init__(
        self,
        nb_landmarks: int,
        sigmas: float | list[float] | torch.Tensor | np.ndarray = 1.0,
        gamma: Optional[float] = None,
        rotation: float | list[float] | torch.Tensor | np.ndarray = 0,
        heatmap_size: tuple[int, ...] = (512, 512),
        learnable: bool = False,
        background: bool = False,
        all_points: bool = False,
        continuous: bool = True,
        na_zero: bool = False,
    ) -> None:
        super(GaussianHeatmapGenerator, self).__init__(
            nb_landmarks,
            sigmas,
            gamma,
            rotation,
            heatmap_size,
            learnable,
            background,
            all_points,
            continuous,
            na_zero=na_zero,
        )

    def heatmap_fun(
        self,
        landmark_t: torch.Tensor,
        coords: torch.Tensor,
        covariance: torch.Tensor,
        gamma: Optional[float],
    ) -> torch.Tensor:
        """Gaussian heatmap function

        Args:
            landmark_t (torch.Tensor): coordinates of the landmark (y, x) or (z, y, x)
            coords (torch.Tensor): coordinates of the pixel (y, x) or (z, y, x)
            covariance (torch.Tensor): covariance matrix (y, x) or (z, y, x)
            gamma (float or None): scaling factor of the heatmap function


        Returns:
            torch.Tensor: value of the gaussian heatmap function for the given pixel
        """
        if len(covariance.shape) == len(landmark_t.shape[: -(self.spatial_dims - 1)]):
            inverse_covariance = (
                torch.inverse(covariance[..., : self.spatial_dims, : self.spatial_dims])
                .unsqueeze(-3)
                .unsqueeze(-3)
            )
        else:
            # multiple of the same landmarks
            inverse_covariance = (
                torch.inverse(covariance[..., : self.spatial_dims, : self.spatial_dims])
                .unsqueeze(-3)
                .unsqueeze(-3)
                .unsqueeze(-3)
            )
        if self.spatial_dims == 3:
            inverse_covariance = inverse_covariance.unsqueeze(-3)
        diff = (landmark_t - coords).unsqueeze(-2)
        # if self.spatial_dims == 3:
        #     assert False, f"diff shape: {diff.shape}" + f"inv shape: {inverse_covariance.shape}"
        if gamma is not None:
            if self.spatial_dims == 2:
                scale = gamma / (
                    (2 * torch.pi) ** (self.spatial_dims / 2)
                    * torch.sqrt(
                        torch.det(covariance[..., : self.spatial_dims, : self.spatial_dims])
                    )
                ).unsqueeze(-1).unsqueeze(-1)
            else:
                scale = gamma / (
                    (2 * torch.pi) ** (self.spatial_dims / 2)
                    * torch.sqrt(
                        torch.det(covariance[..., : self.spatial_dims, : self.spatial_dims])
                    )
                ).unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)

        else:
            scale = 1
        return scale * torch.exp(-0.5 * (diff @ inverse_covariance @ diff.transpose(-2, -1))).view(
            *landmark_t.shape[: -(self.spatial_dims + 1)],
            *coords.shape[-(self.spatial_dims + 1) : -1],
        )


class LaplacianHeatmapGenerator(HeatmapGenerator):
    """
    Laplacian heatmap generator for generating heatmaps from landmarks.

    Args:
        nb_landmarks (int): number of landmarks
        sigmas (float or list[float] or torch.Tensor or np.ndarray): sigmas of the Laplacian heatmap
            function
        gamma (float or None): scaling factor of the Laplacian heatmap function
        rotation (float or list[float] or torch.Tensor or np.ndarray): rotation of the laplacian
            heatmap function
        heatmap_size (tuple[int, int]): size of the heatmap
        learnable (bool): whether the sigmas and rotation are learnable
        background (bool): whether to add a background channel to the heatmap
        all_points (bool): whether to add a channel with the sum of all the landmarks
        continuous (bool): whether to use continuous or discrete landmarks
        na_zero (bool): whether to set the value of the landmarks to zero if they are not available
    """

    def __init__(
        self,
        nb_landmarks: int,
        sigmas: float | list[float] | torch.Tensor | np.ndarray = 1.0,
        gamma: Optional[float] = None,
        rotation: float | list[float] | torch.Tensor | np.ndarray = 0,
        heatmap_size: tuple[int, ...] = (512, 512),
        learnable: bool = False,
        background: bool = False,
        all_points: bool = False,
        continuous: bool = True,
        na_zero: bool = False,
    ) -> None:
        super(LaplacianHeatmapGenerator, self).__init__(
            nb_landmarks,
            sigmas,
            gamma,
            rotation,
            heatmap_size,
            learnable,
            background,
            all_points,
            continuous,
            na_zero=na_zero,
        )
        if self.spatial_dims != 2:
            raise ValueError("Laplacian heatmap generator only works in 2D")

    def heatmap_fun(
        self,
        landmark_t: torch.Tensor,
        coords: torch.Tensor,
        covariance: torch.Tensor,
        gamma: Optional[float] = None,
    ) -> torch.Tensor:
        """Laplacian heatmap function

        Args:
            landmark_t (torch.Tensor): coordinates of the landmark (y, x)
            coords (torch.Tensor): coordinates of the pixel (y, x)
            covariance (torch.Tensor): covariance matrix (y, x)
            gamma (float or None): scaling factor of the heatmap function


        Returns:
            torch.Tensor: value of the gaussian heatmap function for the given pixel
        """
        if len(covariance.shape) == len(landmark_t.shape[:-1]):
            inverse_covariance = (
                torch.inverse(covariance[..., : self.spatial_dims, : self.spatial_dims])
                .unsqueeze(-(self.spatial_dims + 1))
                .unsqueeze(-(self.spatial_dims + 1))
            )
        else:
            # multiple of the same landmarks
            inverse_covariance = (
                torch.inverse(covariance[..., : self.spatial_dims, : self.spatial_dims])
                .unsqueeze(-(self.spatial_dims + 1))
                .unsqueeze(-(self.spatial_dims + 1))
                .unsqueeze(-(self.spatial_dims + 1))
            )
        diff = (landmark_t - coords).unsqueeze(-2)
        if gamma is not None:
            return (
                gamma
                / ((2 / 3) * torch.pi * torch.sqrt(torch.det(covariance[..., :2, :2])))
                .unsqueeze(-1)
                .unsqueeze(-1)
                * torch.exp(
                    -1 * torch.sqrt(3 * (diff @ inverse_covariance @ diff.transpose(-2, -1)))
                ).view(*landmark_t.shape[:-3], *coords.shape[-3:-1])
            )
        return torch.exp(
            -1 * torch.sqrt(3 * (diff @ inverse_covariance @ diff.transpose(-2, -1)))
        ).view(*landmark_t.shape[:-3], *coords.shape[-3:-1])


def from_2by2_to_4by4(affine_matrix: torch.Tensor) -> torch.Tensor:
    """Converts a 2 by 2 affine matrix to a 4 by 4 affine matrix.

    Args:
        affine_matrix (torch.Tensor): 2 by 2 affine matrix

    Returns:
        torch.Tensor: 4 by 4 affine matrix
    """
    assert affine_matrix.shape[-1] == 2
    assert affine_matrix.shape[-2] == 2
    if len(affine_matrix.shape) == 2:
        return torch.cat(
            (
                torch.cat((affine_matrix, torch.zeros((2, 2), device=affine_matrix.device)), dim=1),
                torch.cat(
                    (
                        torch.zeros((2, 2), device=affine_matrix.device),
                        torch.diag(torch.ones(2, device=affine_matrix.device)),
                    ),
                    dim=1,
                ),
            ),
            dim=0,
        )
    if len(affine_matrix.shape) == 3:
        return torch.cat(
            (
                torch.cat(
                    (affine_matrix, torch.zeros(affine_matrix.shape, device=affine_matrix.device)),
                    dim=2,
                ),
                torch.cat(
                    (
                        torch.zeros(affine_matrix.shape, device=affine_matrix.device),
                        torch.diag(torch.ones(2, device=affine_matrix.device)).repeat(
                            affine_matrix.shape[0], 1, 1
                        ),
                    ),
                    dim=2,
                ),
            ),
            dim=1,
        )
    if len(affine_matrix.shape) == 4:
        return from_2by2_to_4by4(affine_matrix.view(-1, 2, 2)).view(
            affine_matrix.shape[0], affine_matrix.shape[1], 4, 4
        )
    raise ValueError("Affine matrix should be of shape (2,2) or (B,2,2) or (B,C,2,2))")


def from_3by3_to_4by4(affine_matrix: torch.Tensor) -> torch.Tensor:
    """Converts a 3 by 3 affine matrix to a 4 by 4 affine matrix.

    Args:
        affine_matrix (torch.Tensor): 3 by 3 affine matrix

    Returns:
        torch.Tensor: 4 by 4 affine matrix
    """
    assert affine_matrix.shape[-1] == 3
    assert affine_matrix.shape[-2] == 3
    if len(affine_matrix.shape) == 2:
        return torch.cat(
            (
                torch.cat((affine_matrix, torch.zeros((3, 1), device=affine_matrix.device)), dim=1),
                torch.cat(
                    (
                        torch.zeros((1, 3), device=affine_matrix.device),
                        torch.diag(torch.ones(1, device=affine_matrix.device)),
                    ),
                    dim=1,
                ),
            ),
            dim=0,
        )
    if len(affine_matrix.shape) == 3:
        return torch.cat(
            (
                torch.cat(
                    (
                        affine_matrix,
                        torch.zeros((affine_matrix.shape[0], 3, 1), device=affine_matrix.device),
                    ),
                    dim=2,
                ),
                torch.cat(
                    (
                        torch.zeros((affine_matrix.shape[0], 1, 3), device=affine_matrix.device),
                        torch.diag(torch.ones(1, device=affine_matrix.device)).repeat(
                            affine_matrix.shape[0], 1, 1
                        ),
                    ),
                    dim=2,
                ),
            ),
            dim=1,
        )
    if len(affine_matrix.shape) == 4:
        return from_3by3_to_4by4(affine_matrix.view(-1, 3, 3)).view(
            affine_matrix.shape[0], affine_matrix.shape[1], 4, 4
        )
    raise ValueError("Affine matrix should be of shape (3,3) or (B,3,3), (B,C,3,3)")



import copy
import torch
from landmarker.heatmap.generator import GaussianHeatmapGenerator
from landmarker.schedulers.adaloss import AdalossScheduler


def test_adaloss_scheduler():
    # create an instance of the scheduler
    scheduler = AdalossScheduler(nb_landmarks=5, rho=0.9, window=3, non_increasing=False)

    heatmap_generator = GaussianHeatmapGenerator(nb_landmarks=5, sigmas=10)

    losses = torch.tensor([0.2, 0.1, 0.2, 0.1, 0.2])
    # call the scheduler with the input losses and sigmas
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 1

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) == 0

    losses = losses*0.9

    # call the scheduler with the same input losses and sigmas again
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 2

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) == 0

    losses = losses*0.9

    # call the scheduler with the same input losses and sigmas again
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 3

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) != 0

    losses = losses*0.9

    # call the scheduler with the same input losses and sigmas again
    prev_sigmas = copy.deepcopy(heatmap_generator.sigmas)
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length and that the prev variances are not zero
    for i in range(5):
        assert len(scheduler.losses[i]) == 3
        assert scheduler.prev_variances[i] != 0

    # check that the sigmas have been updated (they should be smaller)
    assert heatmap_generator.sigmas.shape == prev_sigmas.shape
    for i in range(5):
        assert heatmap_generator.sigmas[i, 0] < prev_sigmas[i, 0]
        assert heatmap_generator.sigmas[i, 1] < prev_sigmas[i, 1]


def test_adaloss_scheduler_non_increasing():
    # create an instance of the scheduler
    scheduler = AdalossScheduler(nb_landmarks=5, rho=0.9, window=3, non_increasing=True)

    heatmap_generator = GaussianHeatmapGenerator(nb_landmarks=5, sigmas=10)

    losses = torch.tensor([0.2, 0.1, 0.2, 0.1, 0.2])
    # call the scheduler with the input losses and sigmas
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 1

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) == 0

    losses = losses*1.1

    # call the scheduler with the same input losses and sigmas again
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 2

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) == 0

    losses = losses*1.1

    # call the scheduler with the same input losses and sigmas again
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length
    for i in range(5):
        assert len(scheduler.losses[i]) == 3

    # check that the prev_variances list has the correct length
    assert len(scheduler.prev_variances) != 0

    losses = losses*1.1

    # call the scheduler with the same input losses and sigmas again
    prev_sigmas = copy.deepcopy(heatmap_generator.sigmas)
    scheduler(losses, heatmap_generator.sigmas)

    # check that the losses deque has the correct length and that the prev variances are not zero
    for i in range(5):
        assert len(scheduler.losses[i]) == 3
        assert scheduler.prev_variances[i] != 0

    # check that the sigmas have been updated (they should be smaller)
    assert heatmap_generator.sigmas.shape == prev_sigmas.shape
    for i in range(5):
        assert heatmap_generator.sigmas[i, 0] == prev_sigmas[i, 0]
        assert heatmap_generator.sigmas[i, 1] == prev_sigmas[i, 1]



"""Tests for utils module."""

import json
import os

import cv2
import numpy as np
import pytest
import torch

from landmarker.utils.utils import (
    all_annotations_to_landmarks,
    all_annotations_to_landmarks_numpy,
    annotation_to_landmark,
    annotation_to_landmark_numpy,
    covert_video_to_frames,
    get_angle,
    get_angle_numpy,
    get_paths,
    pixel_to_unit,
    pixel_to_unit_numpy,
)


@pytest.fixture(scope="session", autouse=True)
def setup_data():
    """Create data for tests."""
    # Create a temporary directory and files for testing
    tmpdir = os.path.join(os.path.dirname(__file__), "tmp")
    os.mkdir(tmpdir)
    test_files = ["file1.txt", "file2.txt", "file3.jpg"]
    for file in test_files:
        with open(os.path.join(tmpdir, file), "a", encoding="utf8") as f:
            f.close()

    # Define test cases
    json_objs = [
        {
            "shapes": [
                {"label": "class1", "points": [[1, 2]]},
                {"label": "class2", "points": [[3, 4]]},
                {"label": "class3", "points": [[5, 6]]},
            ]
        },
        {
            "shapes": [
                {"label": "class1", "points": [[7, 8]]},
                {"label": "class2", "points": [[9, 10]]},
                {"label": "class3", "points": [[11, 12]]},
            ]
        },
    ]

    # Write the test cases to temporary files
    paths = []
    for i, json_obj in enumerate(json_objs):
        path = os.path.join(tmpdir, f"test{i}.json")
        with open(path, "w", encoding="utf8") as f:
            json.dump(json_obj, f)
        paths.append(path)

    yield

    # Remove the temporary directory and files
    for file in test_files:
        os.remove(os.path.join(tmpdir, file))
    for path in paths:
        os.remove(path)

    os.rmdir(tmpdir)


def test_get_paths():
    """Test the get_paths function."""

    # Test if get_paths correctly finds .txt files
    txt_paths = get_paths(os.path.dirname(__file__), "txt")
    assert len(txt_paths) == 2
    assert all(path.endswith(".txt") for path in txt_paths)

    # Test if get_paths correctly finds .jpg files
    jpg_paths = get_paths(os.path.dirname(__file__), "jpg")
    assert len(jpg_paths) == 1
    assert all(path.endswith(".jpg") for path in jpg_paths)


def test_annotation_to_landmark():
    """Test the annotation_to_landmark function."""
    # Define a test case
    json_obj = {
        "shapes": [
            {"label": "class1", "points": [[1, 2]]},
            {"label": "class2", "points": [[3, 4]]},
            {"label": "class3", "points": [[5, 6]]},
        ]
    }
    class_names = ["class1", "class2", "class3"]

    # Call the function with the test case
    landmarks = annotation_to_landmark(json_obj, class_names)

    # Check the result
    expected_landmarks = torch.Tensor([[2, 1], [4, 3], [6, 5]])
    assert torch.allclose(landmarks, expected_landmarks)

    # Test with a class name that is not in the json_obj
    class_names = ["class1", "class4"]
    landmarks = annotation_to_landmark(json_obj, class_names)
    expected_landmarks = torch.Tensor([[2, 1], [float("nan"), float("nan")]])
    assert torch.allclose(landmarks, expected_landmarks, equal_nan=True)


def test_annotation_to_landmark_numpy():
    """Test the annotation_to_landmark function. (Numpy version)"""
    # Define a test case
    json_obj = {
        "shapes": [
            {"label": "class1", "points": [[1, 2]]},
            {"label": "class2", "points": [[3, 4]]},
            {"label": "class3", "points": [[5, 6]]},
        ]
    }
    class_names = ["class1", "class2", "class3"]

    # Call the function with the test case
    landmarks = annotation_to_landmark_numpy(json_obj, class_names)

    # Check the result
    expected_landmarks = np.array([[2, 1], [4, 3], [6, 5]])
    assert np.allclose(landmarks, expected_landmarks)

    # Test with a class name that is not in the json_obj
    class_names = ["class1", "class4"]
    landmarks = annotation_to_landmark_numpy(json_obj, class_names)
    expected_landmarks = np.array([[2, 1], [np.nan, np.nan]])
    assert np.allclose(landmarks, expected_landmarks, equal_nan=True)


def test_all_annotations_to_landmarks():
    """Test the all_annotations_to_landmarks function."""
    tmpdir = os.path.join(os.path.dirname(__file__), "tmp")

    class_names = ["class1", "class2", "class3"]

    # Write the test cases to temporary files
    paths = sorted(list(get_paths(tmpdir, "json")))

    # Call the function with the test cases
    landmarks = all_annotations_to_landmarks(paths, class_names)

    # Check the result
    expected_landmarks = torch.Tensor(
        [
            [[2, 1], [4, 3], [6, 5]],
            [[8, 7], [10, 9], [12, 11]],
        ]
    )
    assert torch.allclose(landmarks, expected_landmarks)


def test_all_annotations_to_landmarks_numpy():
    """Test the all_annotations_to_landmarks function. (Numpy version)"""
    tmpdir = os.path.join(os.path.dirname(__file__), "tmp")

    class_names = ["class1", "class2", "class3"]

    # Write the test cases to temporary files
    paths = sorted(list(get_paths(tmpdir, "json")))

    # Call the function with the test cases
    landmarks = all_annotations_to_landmarks_numpy(paths, class_names)

    # Check the result
    expected_landmarks = np.array(
        [
            [[2, 1], [4, 3], [6, 5]],
            [[8, 7], [10, 9], [12, 11]],
        ]
    )
    assert np.allclose(landmarks, expected_landmarks)


def test_get_angle():
    """Test the get_angle function."""
    # Define test cases
    p1 = torch.Tensor([0, 0])
    p2 = torch.Tensor([1, 0])
    p3 = torch.Tensor([0, 1])

    # Call the function with the test cases
    angle_radial = get_angle(p1, p2, p3, radial=True)
    angle_degrees = get_angle(p1, p2, p3, radial=False)

    # Check the results
    assert torch.allclose(angle_radial, torch.Tensor([np.pi / 2]))
    assert torch.allclose(angle_degrees, torch.Tensor([90]))

    # Test with different points
    p3 = torch.Tensor([1, 1])
    angle_radial = get_angle(p1, p2, p3, radial=True)
    angle_degrees = get_angle(p1, p2, p3, radial=False)

    # Check the results
    assert torch.allclose(angle_radial, torch.Tensor([np.pi / 4]))
    assert torch.allclose(angle_degrees, torch.Tensor([45]))


def test_get_angle_numpy():
    """Test the get_angle function. (Numpy version)"""
    # Define test cases
    p1 = np.array([0, 0])
    p2 = np.array([1, 0])
    p3 = np.array([0, 1])

    # Call the function with the test cases
    angle_radial = get_angle_numpy(p1, p2, p3, radial=True)
    angle_degrees = get_angle_numpy(p1, p2, p3, radial=False)

    # Check the results
    assert np.allclose(angle_radial, np.array([np.pi / 2]))
    assert np.allclose(angle_degrees, np.array([90]))

    # Test with different points
    p3 = np.array([1, 1])
    angle_radial = get_angle_numpy(p1, p2, p3, radial=True)
    angle_degrees = get_angle_numpy(p1, p2, p3, radial=False)

    # Check the results
    assert np.allclose(angle_radial, np.array([np.pi / 4]))
    assert np.allclose(angle_degrees, np.array([45]))


def test_pixel_to_unit_spacing_torch():
    """
    Test the pixel_to_unit function with pixel spacing and torch tensors.
    """
    landmarks = torch.Tensor([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = torch.Tensor([[0.3, 0.1], [0.1, 0.3]])
    dim = (50, 100)
    dim_orig = torch.Tensor([[100, 150], [200, 200]])

    landmarks_resized = (
        landmarks * torch.Tensor(dim).reshape((-1, 1, 2)) / dim_orig.reshape((-1, 1, 2))
    )

    expected_result = landmarks * pixel_spacing.reshape((-1, 1, 2))

    result = pixel_to_unit(landmarks_resized, pixel_spacing, dim, dim_orig)

    assert torch.allclose(result, expected_result)


def test_pixel_to_unit_no_resize_torch():
    """
    Test the pixel_to_unit function without resizing.
    """
    landmarks = torch.Tensor([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = torch.Tensor([[0.3, 0.1], [0.1, 0.3]])

    expected_result = torch.Tensor(
        [
            [[30 * 0.3, 20 * 0.1], [100 * 0.3, 70 * 0.1]],
            [[105 * 0.1, 60 * 0.3], [70 * 0.1, 80 * 0.3]],
        ]
    )

    result = pixel_to_unit(landmarks, pixel_spacing)

    assert torch.allclose(result, expected_result)

    with pytest.raises(ValueError):
        result = pixel_to_unit(landmarks, pixel_spacing, dim=(100, 100))


def test_pixel_to_unit_spacing_numpy():
    """
    Test the pixel_to_unit_numpy function with pixel spacing and numpy arrays.
    """
    landmarks = np.array([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = np.array([[0.3, 0.1], [0.1, 0.3]])
    dim = (50, 100)
    dim_orig = np.array([[100, 150], [200, 200]])

    landmarks_resized = landmarks * np.array(dim).reshape((-1, 1, 2)) / dim_orig.reshape((-1, 1, 2))

    expected_result = landmarks * pixel_spacing.reshape((-1, 1, 2))

    result = pixel_to_unit_numpy(landmarks_resized, pixel_spacing, dim, dim_orig)

    assert np.allclose(result, expected_result)


def test_pixel_to_unit_padding_numpy():
    """
    Test the pixel_to_unit_numpy function with padding and numpy arrays.
    """
    landmarks_orig = np.array([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = None
    dim = (300, 200)
    dim_orig = np.array([[100, 200], [200, 200]])
    padding = np.array([[100, 0], [50, 0]])

    landmarks_padded = landmarks_orig + padding.reshape((-1, 1, 2))
    landmarks = landmarks_padded

    result = pixel_to_unit_numpy(landmarks, pixel_spacing, dim, dim_orig, padding)

    assert np.allclose(result, landmarks_orig)

    landmarks_orig = np.array([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = None
    dim = (150, 100)
    dim_orig = np.array([[100, 200], [200, 200]])
    padding = np.array([[100, 0], [50, 0]])

    landmarks_padded = landmarks_orig + padding.reshape((-1, 1, 2))
    landmarks = landmarks_padded * (
        np.array(dim).reshape((-1, 1, 2)) / (dim_orig + 2 * padding).reshape((-1, 1, 2))
    )

    result = pixel_to_unit_numpy(landmarks, pixel_spacing, dim, dim_orig, padding)

    assert np.allclose(result, landmarks_orig)


def test_pixel_to_unit_no_resize_numpy():
    """
    Test the pixel_to_unit_numpy function without resizing.
    """
    landmarks = np.array([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = np.array([[0.3, 0.1], [0.1, 0.3]])

    expected_result = np.array(
        [
            [[30 * 0.3, 20 * 0.1], [100 * 0.3, 70 * 0.1]],
            [[105 * 0.1, 60 * 0.3], [70 * 0.1, 80 * 0.3]],
        ]
    )

    result = pixel_to_unit_numpy(landmarks, pixel_spacing)

    assert np.allclose(result, expected_result)

    try:
        result = pixel_to_unit_numpy(landmarks, pixel_spacing, dim=(100, 100))
        assert False
    except ValueError:
        assert True


def test_pixel_to_unit_array_dim_numpy():
    """
    Test the pixel_to_unit_numpy function with array dim.
    """
    landmarks = np.array([[[30, 20], [100, 70]], [[105, 60], [70, 80]]])
    pixel_spacing = np.array([[0.3, 0.1], [0.1, 0.3]])
    dim = np.array([(50, 100), (100, 200)])
    dim_orig = np.array([[100, 150], [200, 200]])

    landmarks_resized = landmarks * np.array(dim).reshape((-1, 1, 2)) / dim_orig.reshape((-1, 1, 2))

    expected_result = landmarks * pixel_spacing.reshape((-1, 1, 2))

    result = pixel_to_unit_numpy(landmarks_resized, pixel_spacing, dim, dim_orig)

    assert np.allclose(result, expected_result)


def test_covert_video_to_frames(tmpdir):
    """Test the covert_video_to_frames function."""
    # Create a temporary video file for testing
    video_path = os.path.join(tmpdir, "test.avi")
    height, width, frames = 240, 320, 10
    video = cv2.VideoWriter(video_path, cv2.VideoWriter_fourcc(*"MJPG"), 30, (width, height))
    for _ in range(frames):
        frame = np.random.randint(0, 256, (height, width, 3), dtype=np.uint8)
        video.write(frame)
    video.release()

    # Call the function with the test video
    frames_path = os.path.join(tmpdir, "frames")
    covert_video_to_frames(video_path, frames_path)

    # Check if the correct number of frames were saved
    saved_frames = os.listdir(frames_path)
    assert len(saved_frames) == frames

    # Check if the frames were saved with the correct names
    for i in range(1, frames + 1):
        assert f"img_{str(i).zfill(6)}.png" in saved_frames
